import requests


class HttpClient:
    """Client for handling requests to Tonic instance

    Parameters
    ----------
    base_url : str
        URL to Tonic instance
    api_key : str
        The API token associated with your workspace
    """

    def __init__(self, base_url: str, api_key: str):
        self.base_url = base_url
        self.headers = {"Authorization": api_key}

    def http_get(self, url: str, params: dict={}):
        """Make a get request.

        Parameters
        ----------
        url : str
            URL to make get request. Is appended to self.base_url.
        params: dict
            Passed as the params parameter of the requests.get request.
        
        """
        res = requests.get(self.base_url + url, params=params, headers=self.headers, verify=False)
        res.raise_for_status()
        return res.json()

    def http_post(self, url, params={}, data={}):
        """Make a post request.

        Parameters
        ----------
        url : str
            URL to make the post request. Is appended to self.base_url.
        params: dict
            Passed as the params parameter of the requests.post request.
        data: dict
            Passed as the data parameter of the requests.post request.
        """
        res = requests.post(
            self.base_url + url, params=params, json=data, headers=self.headers, verify=False
        )
        res.raise_for_status()
        return res.json()


class Model:
    """
    Class to store necessary metadata for a given model.

    Parameters
    ----------
    id : str
        Model id, comes from the workspace json/dictionary.

    model_json : dict
        Model json, comes from the workspace json/dictionary.
    """
    def __init__(self, id: str, model_json: dict):
        self.id = id
        self.name = model_json["name"]
        self.query = model_json["query"]
        self.parameters = model_json["nnModelConfig"]
        self.encodings = model_json["columnEncodings"]
        self.model_type = self.parameters["modelType"]
        self.conditions = model_json.get("conditionColumns", [])
        self.has_conditions = (self.model_type == 'RNN_VAE') and len(self.conditions) > 0

    def describe(self):
        """Print model metadata - id, query, parameters (model config), encodings.
        """
        print("Model: " + self.name + " [" + self.id + "]")
        print("Query: " + self.query)
        print("Parameters: " + str(self.parameters))
        print("Column Encodings: " + str(self.encodings))
        if self.has_conditions:
            print("Columns conditioned on: " + str(self.conditions))



from tonic_api.classes.httpclient import HttpClient
from tonic_api.services.workspaces import WorkspaceService
from tonic_api.classes.workspace import Workspace


class TonicApi:
    '''Wrapper class for invoking Tonic API
    
    Parameters
    ----------
    base_url : str
        The url to your Tonic instance. Do not include trailing backslashes
    api_key : str
        Your api token

    Examples
    --------
    >>> TonicApi("http://localhost:3000", "your_api_key")
    '''
    def __init__(self, base_url : str, api_key: str):
        client = HttpClient(base_url, api_key)
        self.workspace_service = WorkspaceService(client)

    def get_workspace(self, workspace_id : str) -> Workspace:
        '''Get instance of Workspace class with specified workspace_id.

        Parameters
        ----------
        workspace_id : str
            The id for your workspace

        Returns
        -------
        Workspace

        Examples
        --------
        >>> workspace = tonic.get_workspace("11529c2f-e9a7-6c12-4025-c2cdc0dcb6fa")
        '''
        return self.workspace_service.get_workspace(workspace_id)



import pandas as pd
import requests
from typing import List, Union

from tonic_api.services.models import ModelService
from tonic_api.classes.model import Model
from tonic_api.classes.httpclient import HttpClient


class TrainedModel:
    """Wrapper class for accessing trained models.

    Attributes
    ----------
    id : str
        id of model
    job_id : str
        id of training job
    workspace_id : str
        id of workspace
    model :  Model
        the Tonic Model object
    model_service : ModelService
        service for accessing Tonic Model
    """

    def __init__(
        self, id: str, job_id: str, workspace_id: str, model: Model, client: HttpClient
    ):
        self.id = id
        self.job_id = job_id
        self.workspace_id = workspace_id
        self.model = model
        self.client = client
        self.model_service = ModelService(client)

    def sample(self, num_rows: int=1) -> pd.DataFrame:
        """Generates synthetic samples from trained_model.

        Parameters
        ----------
        num_rows : int, optional
            Number of rows to generate.

        Returns
        -------
        pandas.DataFrame
            A dataframe of synthetic data.

        Examples
        --------
        >>> synth_df = model.sample(1000)
        """

        num_rows = self.__validate_sample_input(num_rows)
        res = self.model_service.sample(self.id, num_rows)
        return self.__convert_to_df(res)

    def conditional_sample(
        self, conditions: Union[List[dict], pd.DataFrame]
    ) -> pd.DataFrame:
        """Generates synthetic samples from from conditional model, or throws an error
        if the model is not conditional.

        The synthetic samples will be generated based on the specified conditions, with
        one complete sequence of samples for each condition. Note that the sequences
        will be returned in the same order as the conditions, as the conditions will
        be present in the output dataframe.
        
        The conditions parameter can be a list of dictionaries, or a pandas DataFrame. 
        In either case, the values of all condition columns must be specified.

        Parameters
        ----------
        conditions : List[dict] or pd.DataFrame
            List or dataframe of conditions. If a list, each list item should be a 
            dictionary with keys for each condition column and values for the condition. 
            If a dataframe, the extra columns will be dropped and the remaining columns 
            will be used as conditions.

        Returns
        -------
        pandas.DataFrame
            A dataframe of synthetic data.
        """
        if not self.model.has_conditions:
            raise ValueError("Model does not have conditions")

        if isinstance(conditions, pd.DataFrame):
            columns = conditions.columns.tolist()
            for col in columns:
                if col not in self.model.conditions:
                    print("Ignoring column " + col + " not in conditions")
            conditions = conditions[self.model.conditions].to_dict("records")

        try:
            res = self.model_service.conditional_sample(self.id, conditions)
            return self.__convert_to_df(res)
        except requests.exceptions.RequestException as err:
            print('ERROR: {}'.format(err.response.text))

    def sample_source(self, num_rows: int=1):
        """Generates samples from source data.

        Parameters
        ----------
        num_rows : int, optional
            Number of rows to generate.

        Returns
        -------
        pandas.DataFrame
            A dataframe of real data.

        Examples
        --------
        >>> real_df = model.sample_source(1000)
        """
        num_rows = self.__validate_sample_input(num_rows)

        res = self.model_service.sample_source(
            self.workspace_id, self.model.query, num_rows
        )
        columns = res["columns"]
        if len(columns) == 0:
            raise Exception("No data returned from source database")

        n_rows_returned = len(columns[0]["data"])
        converted_res = [{} for _ in range(n_rows_returned)]

        if n_rows_returned < num_rows:
            print(
                "Not enough rows in source destination to sample, limiting to "
                + str(n_rows_returned)
                + " rows."
            )

        for col in columns:
            data = col["data"]
            for idx, val in enumerate(data):
                converted_res[idx][col["columnName"]] = val
        return self.__convert_to_df(converted_res)

    def get_numeric_columns(self) -> List[str]:
        """Returns list of columsn with numeric encodings.
        
        Returns
        -------
        List[str]
        """

        return [
            col
            for (col, encoding) in self.model.encodings.items()
            if encoding == "Numeric"
        ]

    def get_categorical_columns(self) -> List[str]:
        """Returns list of columns with categorical encodings.
        
        Returns
        -------
        List[str]
        """
        return [
            col
            for (col, encoding) in self.model.encodings.items()
            if encoding == "Categorical"
        ]

    def get_condition_columns(self) -> List[str]:
        conditions = self.model.conditions
        if conditions is not None or len(conditions) > 0:
            return conditions
        print("No conditions found")

    def __convert_to_df(self, sample_response):
        schema = self.__get_schema()
        df = pd.DataFrame(sample_response)
        df = self.__conform_df_to_schema(df, schema)
        return df

    def __get_schema(self):
        schema = self.model_service.get_schema(self.workspace_id, self.model.query)
        ordered_schema = self.__convert_schema_to_ordered_col_list(schema)
        return ordered_schema

    def __convert_schema_to_ordered_col_list(self, schema):
        return [
            obj["columnName"]
            for obj in sorted(schema, key=lambda v: v["ordinalPosition"])
        ]

    def __conform_df_to_schema(self, df, schema):
        return df.reindex(schema, axis=1)
    
    def __validate_sample_input(self, num_rows):
        try:
            num_rows = int(num_rows)
            assert num_rows > 0
        except:
            raise ValueError(f"cannot sample, {num_rows} should be an positive integer")
        return num_rows


    def describe(self):
        """Prints description of trained model."""
        print("Trained Model: [" + self.id + "]")
        print("Job ID: " + self.job_id)
        print("Workspace ID: " + self.workspace_id)
        print("Model: ")
        self.model.describe()



from typing import List
from ..classes.httpclient import HttpClient

class ModelService:
    def __init__(self, client: HttpClient):
        self.client = client

    def get_trained_model(self, job_id: str):
        return self.client.http_get("/api/modeltraining/models/" + job_id)
    
    def get_most_recent_trained_model(self, workspace_id: str, model_id: str):
        params = {"workspaceId": workspace_id, "modelId": model_id}
        return self.client.http_get(
            "/api/modeltraining/most_recent_trained_model", params=params
        )
    
    def sample(self, model_training_id: str, num_rows: int):
        params = {"numRows": num_rows} if num_rows is not None else {}
        return self.client.http_get(
            "/api/modeltraining/sample/" + model_training_id, params=params
        )
    
    def conditional_sample(self, model_training_id: str, conditions_list: List[dict]):
        return self.client.http_post(
            "/api/modeltraining/sample_sequences_conditionally/" + model_training_id,
            data=conditions_list
        )

    def sample_source(self, workspace_id, query, num_rows):
        data = {"workspaceId": workspace_id, "query": query, "numRows": num_rows}
        return self.client.http_post("/api/model/get_preview_raw", data=data)

    def get_schema(self, workspace_id, query):
        data = {"workspaceId": workspace_id, "query": query}
        return self.client.http_post("/api/model/get_schema_of_query", data=data)
    
    def get_job_status(self, job_id: str):
        return self.client.http_get("/api/job/" + job_id)
    
    def start_training(self, workspace_id: str, model_id: str):
        params = {"workspaceId": workspace_id, "modelId": model_id}
        return self.client.http_post(
            "/api/modeltraining/train", params=params
        )


import re

class TrainingJobStatus:
    """Keeps track of information about the status of a training job.

    Parameters
    ----------
    job_json : dict
    """
    def __init__(self, job_json: dict):
        self.state = job_json["status"]
        self.error = job_json["errorMessages"] if "errorMessages" in job_json else None
        self.tasks = job_json["tasks"] if "tasks" in job_json else None

    def current_epoch_progress(self) -> dict:
        """If the job is currently training, returns the epoch the training job is on.
        
        Returns
        -------
        dict
        """
        epoch_metadata = {}
        if self.tasks is not None:
            epoch_task = [
                task
                for task in self.tasks
                if "Training AI Synthesizer Model" in task["action"]
            ]
            if len(epoch_task) > 1:
                raise ValueError(
                    f"there are {len(epoch_task)} training tasks when there should be 1"
                )
            if len(epoch_task) == 1:
                task = epoch_task[0]
                epoch_metadata["model_name"] = task["action"].split()[-1]
                epoch_metadata["steps_completed"] = str(task["stepsCompleted"])
                epoch_metadata["total_steps"] = str(task["totalSteps"])
                return epoch_metadata
        return None

    def describe(self):
        """Print job status."""
        print("Job status: " + self.state)
        if self.error is not None:
            print('Job failed due to: "' + self.error + '"')



from ..classes.workspace import Workspace
from ..classes.model import Model


class WorkspaceService:
    def __init__(self, client):
        self.client = client

    def get_workspace(self, workspace_id):
        workspace = self.client.http_get("/api/workspace/" + workspace_id)
        return Workspace(
            workspace["id"],
            workspace["workspaceName"],
            [Model(id, model) for (id, model) in workspace["models"].items()],
            self.client,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tonic_api', 'tonic_api.classes', 'tonic_api.services']

package_data = \
{'': ['*']}

install_requires = \
['more-itertools>=8.6.0,<9.0.0',
 'pandas>=1.0.0,<2.0.0',
 'requests>=2.23.0,<3.0.0']

setup_kwargs = {
    'name': 'tonic-api',
    'version': '2.1.3',
    'description': 'Wrappers around the Tonic.ai API',
    'long_description': '# Overview\nThis library contains useful wrappers around the Tonic.ai API.\n\n## Usage\n\nInstantiate the API wrapper using the following code:\n\n```\nfrom tonic_api.api import TonicApi\n\n# Do not include trailing backslash in TONIC_URL\napi = TonicApi(TONIC_URL, API_KEY)\n```\n\nOnce instantiated, the following endpoints are available for consumption. Note that available endpoints and response types are limited. Available fields may be severely limited compared to the current Tonic API.\n\n```\nTonicApi:\n    get_workspace(workspace_id) => Workspace\n\nWorkspace:\n    id => string\n    name => string\n    models => Model[]\n\n    train(model_id) => new job ID\n    get_trained_model_by_training_job_id(job_id) => TrainedModel (or None if training is not done)\n    get_training_status_by_training_job_id(job_id) => TrainingJobStatus\n    get_most_recent_trained_model_by_model_id => TrainedModel\n\n    describe() => debugger helper for printing fields\n\nModel:\n    id => string\n    name => string\n    query => string\n    parameters => {}\n    encodings => {}\n\n    describe() => debugger helper for printing fields\n\nTrainedModel:\n    id => string\n    job_id => string\n    model => Model\n\n    sample(num_rows) => pandas DataFrame (defaults to 1 row if num_rows not provided)\n    sample_source(num_rows) => pandas DataFrame (defaults to 1 row if num_rows not provided). Note: NOT randomized. Upper limit is limited to row count in source.\n\n    get_numeric_columns() => string[]\n    get_categorical_columns() => string[]\n\n    describe() => debugger helper for printing fields\n\nTrainingJobStatus:\n    state => string\n    error => string[] or None\n    tasks => {}\n\n    current_epoch_progress() => dict or None\n\n    describe() => debugger helper for printing fields\n```\n',
    'author': 'Eric Timmerman',
    'author_email': 'eric@tonic.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.tonic.ai/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import time
from datetime import datetime
from more_itertools import first_true
from ..services.models import ModelService
from typing import List, Optional, Tuple
from ..classes.model import Model
from ..classes.httpclient import HttpClient
from ..classes.trained_model import TrainedModel
from .training_job_status import TrainingJobStatus


class Workspace:
    """Class to represent and access a Tonic workspace.

    Parameters
    ----------
    id: str
        Workspace id.
    name: str
        Workspace name.
    models: List[Model]
        List of models in the workspace.
    client: HttpClient
        The http client to use.
    """
    def __init__(self, id: str, name: str, models: List[Model], client: HttpClient):
        self.id = id
        self.name = name
        self.models = models
        self.client = client
        self.model_service = ModelService(client)

    # this should take a model id as parameter
    def train(self, model_id: str) -> str:
        """Train a model with the current configuration for the model_id.
        """
        res = self.model_service.start_training(self.id, model_id)
        return res["id"]

    def get_trained_model_by_training_job_id(
        self, job_id: str
    ) -> Optional[TrainedModel]:
        job = self.model_service.get_job_status(job_id)
        status = TrainingJobStatus(job)
        if status.state != "Completed":
            status_messages = ["This training job is not complete."]
            self.__print_with_timestamp(status_messages)
            self.training_job_status(job_id)
        else:
            trained_model = self.model_service.get_trained_model(job_id)
            return self.__convert_to_trained_model(trained_model)

    def get_training_status_by_training_job_id(self, job_id: str) -> TrainingJobStatus:
        job = self.model_service.get_job_status(job_id)
        return TrainingJobStatus(job)
    
    def training_job_status(self, job_id: str):
        job = self.model_service.get_job_status(job_id)
        status = TrainingJobStatus(job)
        status_messages, _ = self.__handle_job_status(status, {})
        self.__print_with_timestamp(status_messages)
    
    def tail_training_job_status(self, job_id: str):
        """Print training status every ~5 seconds until the training job is done.

        Examples
        --------
        >>> training_job.tail_training_status()
        [07/06/22 15:27:50] Training status: Running your_model (905/3000 epochs completed)
        [07/06/22 15:28:00] Training status: Running your_model (906/3000 epochs completed)
        [07/06/22 15:28:16] Training status: Running your_model (907/3000 epochs completed)
        [07/06/22 15:28:32] Training status: Running your_model (908/3000 epochs completed)
        """
        status = None
        last_epoch_metadata = {}
        while status is None or status.state != "Completed":
            status = self.get_training_status_by_training_job_id(job_id)

            status_messages, last_epoch_metadata = self.__handle_job_status(
                status, last_epoch_metadata
            )

            if status.state in ["Completed", "Failed", "Canceled"]:
                break

            self.__print_with_timestamp(status_messages)

            time.sleep(5)

        # Last message(s)
        self.__print_with_timestamp(status_messages)
    
    def get_most_recent_trained_model_by_model_id(
        self, model_id: str
    ) -> Optional[TrainedModel]:
        """"Returns the most recently trained model for a given model id.
        """
        trained_model = self.model_service.get_most_recent_trained_model(
            self.id,  model_id
        )
        return self.__convert_to_trained_model(trained_model)

    def __convert_to_trained_model(self, model_json):
        return TrainedModel(
            model_json["modelTrainingId"],
            model_json["jobId"],
            self.id,
            Model(model_json["modelId"], model_json["model"]),
            self.client,
        )

    def __handle_epoch_updates(
        self,
        status: TrainingJobStatus,
        last_epoch_metadata: dict,
    ):
        status_messages = []
        epoch_metadata = status.current_epoch_progress()
        if epoch_metadata is not None:
            # Do not print epoch state when completed and we've never printed before
            if last_epoch_metadata == {} and status.state == "Completed":
                pass
            # Print epoch state when (not completed and we've never printed before)
            # or epoch updated
            elif last_epoch_metadata == {} or (
                epoch_metadata["steps_completed"]
                != last_epoch_metadata["steps_completed"]
            ):
                status_messages.append(
                    "Training status: Running "
                    + epoch_metadata["model_name"]
                    + " ("
                    + epoch_metadata["steps_completed"]
                    + "/"
                    + epoch_metadata["total_steps"]
                    + " epochs completed)"
                )
                last_epoch_metadata = epoch_metadata
        else:
            status_messages = [
                "Training status: Running (no training tasks reported yet)"
            ]

        return status_messages, last_epoch_metadata

    def __print_with_timestamp(self, messages):
        now = datetime.now().strftime("%D %H:%M:%S")
        for message in messages:
            print("[" + now + "] " + message)
    
    def __handle_job_status(
        self, status: TrainingJobStatus, last_epoch_metadata: dict
    ) -> Tuple[List[str], dict]:
        if status.state == "Completed":
            # Show the final epoch training status on completion
            (
                status_messages,
                last_epoch_metadata,
            ) = self.__handle_epoch_updates(status, last_epoch_metadata)
            status_messages.append("Training completed")
        elif status.state == "Failed":
            status_messages = ["Training failed: " + status.error]
        elif status.state == "Canceled":
            status_messages = ["Training was canceled by user"]
        elif status.state == "Running":
            (
                status_messages,
                last_epoch_metadata,
            ) = self.__handle_epoch_updates(status, last_epoch_metadata)
        else:
            status_messages = ["Training status: " + status.state]
        return status_messages, last_epoch_metadata

    def describe(self):
        """Print the workspace name, id, and number of models.
        
        Examples
        --------
        >>> workspace.describe()
        Workspace: your_workspace_name [b5f37f37-2665-a921-64fc-0450e64f0f51]
        Number of Models: 2
        """
        print("Workspace: " + self.name + " [" + self.id + "]")
        print("Number of Models: " + str(len(self.models)))
